"""
This module contains all the classes and functions that make HTTP requests to the Mail.TM API (client- less).
It is divided into two main classes:

- get
    which is a synchronous class
- xget
    which is an asynchronous class.

Each class has methods that correspond to the API endpoints.

- These methods will take the necessary parameters and make a request to the corresponding endpoint.
- The methods will return the response of the API in the form of a type hinteded object.
- If the method returns `None`, it means that the request failed and the error is not explicitly handled by the method.

The methods of `xget` return `None` and the error is handled by the method itself, so you don't need to handle it.

All the methods in this module make requests to the API.

The methods in this module are also documented here.
"""

from __future__ import annotations

__all__ = ["get", "xget"]


import requests
import aiohttp
import msgspec
import typing as t
import urllib.parse
from ..abc.generic import Token
from ..abc.modals import Account, Domain, DomainPageView
from ..core.methods import AccountMethods, DomainMethods
from ..core.errors import (
    AccountTokenInvalid,
    MissingArgument,
    EntityNotFound,
    MethodNotAllowed,
    RefusedToProcess,
    EntityNotProcessable,
    RatelimitError,
)


class get:
    """
    A synchronous implementation which handles data client-less (without making a session).

    Synchronous
    -----------
    Gets an out-of-client class that helps to fetch data regardless of account token.
    But this won't include all the data that is available in the API. It only pulls the
    data that is avilable without authentication using headers.
    """

    def __init__(self) -> None:
        self._base_url = "https://api.mail.tm"

    def _interact(
        self,
        method: t.Literal["GET", "POST", "DELETE", "PATCH"],
        url: str,
        body: t.Optional[t.Any] = None,
        params: t.Optional[t.Dict[str, str]] = None,
    ) -> t.Optional[bytes]:
        """
        Interact with the API using methods, and API slug.

        Parameters
        ----------
        method: Literal["GET", "POST", "DELETE", "PATCH"]
            The method to use to interact with the API.
        url: str
            The API slug to interact with.
        body: Optional[Any]
            The body of the request.
        params: Optional[Dict[str, str]]
            The query parameters of the request.

        Returns
        -------
        Optional[bytes]
            The response from the API.
        """
        if method not in ["GET", "POST", "DELETE", "PATCH"]:
            raise ValueError("Invalid HTTP method")

        if method == "GET":
            resp = requests.get(url=url, params=params, json=body)
        elif method == "POST":
            resp = requests.post(url=url, params=params, json=body)
        elif method == "DELETE":
            resp = requests.delete(url=url, params=params, json=body)
        elif method == "PATCH":
            resp = requests.patch(url=url, params=params, json=body)

        if str(resp.status_code).startswith("20"):
            return resp.content
        elif resp.status_code == 400:
            raise MissingArgument(
                "Something in your payload is missing! Or, the payload isn't there at all."
            )
        elif resp.status_code == 401:
            raise AccountTokenInvalid(
                "Your token isn't correct (Or the headers hasn't a token at all!). Remember, every request (Except POST /accounts and POST /token) should be authenticated with a Bearer token!"
            )
        elif resp.status_code == 404:
            raise EntityNotFound(
                "You're trying to access an account that doesn't exist? Or maybe reading a non-existing message? Go check that!"
            )
        elif resp.status_code == 405:
            raise MethodNotAllowed(
                "Maybe you're trying to GET a /token or POST a /messages. Check the path you're trying to make a request to and check if the method is the correct one."
            )
        elif resp.status_code == 418:
            raise RefusedToProcess(
                "Server is a teapot. And refused to process your request at the moment. Kindly contact the developers for further details."
            )
        elif resp.status_code == 422:
            raise EntityNotProcessable(
                "Some went wrong on your payload. Like, the username of the address while creating the account isn't long enough, or, the account's domain isn't correct. Things like that."
            )
        elif resp.status_code == 429:
            raise RatelimitError(
                "You exceeded the limit of 8 requests per second! Try delaying the request by one second!"
            )
        else:
            raise ValueError(f"Unknown Error\n TB:\n{resp.text}")

    def create_account(
        self, address: str, password: str
    ) -> t.Optional[Account]:
        """
        Creates an account.

        Parameters
        ----------
        address: str
            The email address of the new account.
        password: str
            The password for the new account.

        Returns
        -------
        Optional[Account]
            The newly created account object if successful, None otherwise.
        """
        body = {"address": f"{address}", "password": f"{password}"}
        resp = self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url, AccountMethods.CREATE_ACCOUNT
            ),
            body=body,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    def get_account(self, account_id: str) -> t.Optional[Account]:
        """
        Gets an account using ID.

        Parameters
        ----------
        account_id: str
            The ID of the account to get.

        Returns
        -------
        Optional[Account]
            The account object if successful, None otherwise.
        """
        resp = self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url,
                AccountMethods.GET_ACCOUNT_BY_ID.format(id=f"{account_id}"),
            ),
            params={"id": f"{account_id}"},
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    def delete_account(self, account_id: str) -> bool:
        """
        Deletes an account using ID.

        Parameters
        ----------
        account_id: str
            The ID of the account to delete.

        Returns
        -------
        bool
            True if successful, False otherwise.
        """
        resp = requests.delete(
            url=urllib.parse.urljoin(
                self._base_url,
                AccountMethods.DELETE_ACCOUNT_BY_ID.format(id=f"{account_id}"),
            ),
            params={"id": f"{account_id}"},
        )
        return resp.status_code == 204

    def get_account_token(
        self, account_address: str, account_password: str
    ) -> t.Optional[Token]:
        """
        Get an account token which is used by the clients.

        Parameters
        ----------
        account_address: str
            The email address of the account.
        account_password: str
            The password of the account.

        Returns
        -------
        Optional[Token]
            The account token if successful, None otherwise.
        """
        body = {
            "address": f"{account_address}",
            "password": f"{account_password}",
        }
        resp = self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url, AccountMethods.GET_ACCOUNT_TOKEN
            ),
            body=body,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Token, strict=False)
        else:
            return None

    def get_domain(self, domain_id: str) -> t.Optional[Domain]:
        """
        Get a specific domain with ID.

        Parameters
        ----------
        domain_id: str
            The ID of the domain to get.

        Returns
        -------
        Optional[Domain]
            The domain with the ID provided. If not found, returns None.
        """
        resp = self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url,
                DomainMethods.GET_DOMAIN_BY_ID.format(id=f"{domain_id}"),
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Domain, strict=False)
        else:
            return None

    def get_domains(self) -> t.Optional[DomainPageView]:
        """
        Get all domains.

        Returns
        -------
        Optional[DomainPageView]
            The domain page view if successful, None otherwise.
        """
        resp = self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url, DomainMethods.GET_ALL_DOMAINS
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=DomainPageView, strict=False)
        else:
            return None


class xget:
    """
    An asynchronous implementation which handles data client-less (without making a session).

    Asynchronous
    ------------
    Gets an out-of-client class that helps to fetch data regardless of account token.
    But this won't include all the data that is available in the API. It only pulls the
    data that is avilable without authentication using headers.
    """

    def __init__(self) -> None:
        self._base_url = "https://api.mail.tm"

    async def _interact(
        self,
        method: t.Literal["GET", "POST", "DELETE", "PATCH"],
        url: str,
        body: t.Optional[t.Any] = None,
        params: t.Optional[t.Dict[str, str]] = None,
    ) -> t.Optional[bytes]:
        if method not in ["GET", "POST", "DELETE", "PATCH"]:
            raise ValueError("Invalid HTTP method")
        try:
            async with aiohttp.request(
                method, url, params=params, json=body
            ) as resp:
                result = resp
                if str(result.status).startswith("20"):
                    return await result.read()
                elif result.status == 400:
                    raise MissingArgument(
                        "Something in your payload is missing! Or, the payload isn't there at all."
                    )
                elif result.status == 401:
                    raise AccountTokenInvalid(
                        "Your token isn't correct (Or the headers hasn't a token at all!). Remember, every request (Except POST /accounts and POST /token) should be authenticated with a Bearer token!"
                    )
                elif result.status == 404:
                    raise EntityNotFound(
                        "You're trying to access an account that doesn't exist? Or maybe reading a non-existing message? Go check that!"
                    )
                elif result.status == 405:
                    raise MethodNotAllowed(
                        "Maybe you're trying to GET a /token or POST a /messages. Check the path you're trying to make a request to and check if the method is the correct one."
                    )
                elif result.status == 418:
                    raise RefusedToProcess(
                        "Server is a teapot. And refused to process your request at the moment. Kindly contact the developers for further details."
                    )
                elif result.status == 422:
                    raise EntityNotProcessable(
                        "Some went wrong on your payload. Like, the username of the address while creating the account isn't long enough, or, the account's domain isn't correct. Things like that."
                    )
                elif result.status == 429:
                    raise RatelimitError(
                        "You exceeded the limit of 8 requests per second! Try delaying the request by one second!"
                    )
                else:
                    raise ValueError(
                        f"Unknown Error\nPayload: {(await result.read()).decode()}"
                    )
        except Exception as e:
            print(f"{str(e)}")
            return None

    async def create_account(
        self, address: str, password: str
    ) -> t.Optional[Account]:
        """
        Creates an account.

        Parameters
        ----------
        address: str
            The email address of the new account.
        password: str
            The password for the new account.

        Returns
        -------
        Optional[Account]
            The newly created account object if successful, None otherwise.
        """
        body = {"address": f"{address}", "password": f"{password}"}
        resp = await self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url, AccountMethods.CREATE_ACCOUNT
            ),
            body=body,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    async def get_account(self, account_id: str) -> t.Optional[Account]:
        """
        Gets an account using ID.

        Parameters
        ----------
        account_id: str
            The ID of the account to get.

        Returns
        -------
        Optional[Account]
            The account object if successful, None otherwise.
        """
        resp = await self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url,
                AccountMethods.GET_ACCOUNT_BY_ID.format(id=account_id),
            ),
            params={"id": f"{account_id}"},
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    async def delete_account(self, account_id: str) -> None:
        """
        Deletes an account using ID.

        Parameters
        ----------
        account_id: str
            The ID of the account to delete.

        Returns
        -------
        None
        """
        await self._interact(
            method="DELETE",
            url=urllib.parse.urljoin(
                self._base_url,
                AccountMethods.DELETE_ACCOUNT_BY_ID.format(id=account_id),
            ),
            params={"id": f"{account_id}"},
        )

    async def get_account_token(
        self, account_address: str, account_password: str
    ) -> t.Optional[Token]:
        """
        Get an account token which is used by the clients.

        Parameters
        ----------
        account_address: str
            The email address of the account.
        account_password: str
            The password of the account.

        Returns
        -------
        Optional[Token]
            The account token if successful, None otherwise.
        """
        body = {
            "address": f"{account_address}",
            "password": f"{account_password}",
        }
        resp = await self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url, AccountMethods.GET_ACCOUNT_TOKEN
            ),
            body=body,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Token, strict=False)
        else:
            return None

    async def get_domain(self, domain_id: str) -> t.Optional[Domain]:
        """
        Get a specific domain with ID.

        Parameters
        ----------
        domain_id: str
            The ID of the domain to get.

        Returns
        -------
        Optional[Domain]
            The domain with the ID provided. If not found, returns None.
        """
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url,
                DomainMethods.GET_DOMAIN_BY_ID.format(id=domain_id),
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Domain, strict=False)
        else:
            return None

    async def get_domains(self) -> t.Optional[DomainPageView]:
        """
        Get all domains.

        Returns
        -------
        Optional[DomainPageView]
            The domain page view if successful, None otherwise.
        """
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url, DomainMethods.GET_ALL_DOMAINS
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=DomainPageView, strict=False)
        else:
            return None



import sys
import typing as t
from enum import Enum

from ..abc.modals import Account, Message, Domain


class CacheType(Enum):
    """
    Class that represents different cache data-types.
    """

    NEW_ACCOUNTS = "account"
    """
    All the new accounts that have been cached ever since the server started.
    `mail_client` operations are never cached.
    """
    OLD_MESSAGE = "old_message"
    """
    All the old messages that have been cached ever since the server started.
    `mail_client` operations are never cached.
    """
    NEW_MESSAGE = "new_message"
    """
    All the new messages that have been cached ever since the server started.
    `mail_client` operations are never cached.
    """
    DOMAIN = "domain"
    """
    All the domain changes that have taken place.
    """


class InternalCache:
    """
    A class to manage an internal cache for different types of data.
    """

    def __init__(self) -> None:
        self.internal_memory_map: t.Dict[
            CacheType,
            t.Union[t.List[Message], t.List[Account], t.List[Domain]],
        ] = {}

    def build_cache(self) -> None:
        """
        Builds the initial cache structure with empty lists for different cache types.
        """
        self.internal_memory_map.update(
            {
                CacheType.DOMAIN: [],
                CacheType.NEW_ACCOUNTS: [],
                CacheType.NEW_MESSAGE: [],
                CacheType.OLD_MESSAGE: [],
            }
        )

    def get_old_messages(self) -> t.Optional[t.List[Message]]:
        """
        Retrieves a list of old messages from the cache.

        Returns:
            Optional[List[Message]]: A list of old messages if available, or None if no old messages are found.
        """

        retunable_list: t.List[Message] = []
        base = self.internal_memory_map.get(CacheType.OLD_MESSAGE)
        if base is None:
            return None
        for item in base:
            if isinstance(item, Message):
                retunable_list.append(item)
        return retunable_list

    def get_new_messages(self) -> t.Optional[t.List[Message]]:
        """
        Retrieves a list of new messages from the cache.

        Returns:
            Optional[List[Message]]: A list of new messages if available, or None if no new messages are found.
        """

        retunable_list: t.List[Message] = []
        base = self.internal_memory_map.get(CacheType.NEW_MESSAGE)
        if base is None:
            return None
        for item in base:
            if isinstance(item, Message):
                retunable_list.append(item)
        return retunable_list

    def get_new_accounts(self) -> t.Optional[t.List[Account]]:
        """
        Retrieves a list of new accounts from the cache.

        Returns:
            Optional[List[Account]]: A list of new accounts if available, or None if no new accounts are found.
        """

        retunable_list: t.List[Account] = []
        base = self.internal_memory_map.get(CacheType.NEW_ACCOUNTS)
        if base is None:
            return None
        for item in base:
            if isinstance(item, Account):
                retunable_list.append(item)
        return retunable_list

    def get_domain_cache(self) -> t.Optional[t.List[Domain]]:
        """
        Retrieves the domain cache from the internal memory map.

        Returns:
            Optional[List[Domain]]: A list of domains if the cache is available,
            otherwise None.
        """

        retunable_list: t.List[Domain] = []
        base = self.internal_memory_map.get(CacheType.DOMAIN)
        if base is None:
            return None
        for item in base:
            if isinstance(item, Domain):
                retunable_list.append(item)
        return retunable_list

    def reset_cache(self) -> None:
        """
        Resets the internal cache of the InternalCache object.

        This function clears all the existing data in the internal memory map of the InternalCache object and replaces it with empty lists for each CacheType.

        Returns:
            None
        """
        self.internal_memory_map = {
            CacheType.DOMAIN: [],
            CacheType.NEW_ACCOUNTS: [],
            CacheType.NEW_MESSAGE: [],
            CacheType.OLD_MESSAGE: [],
        }

    def add_item_to_cache(
        self, cache_type: CacheType, item: t.Union[Message, Account, Domain]
    ) -> None:
        """
        Adds an item to the internal cache of the InternalCache object.

        Args:
            cache_type (CacheType): The type of cache to add the item to.
            item (Union[Message, Account, Domain]): The item to add to the cache.

        Returns:
            None
        """
        self.internal_memory_map[cache_type].append(item)  # type: ignore

    def get_cache_size(self) -> int:
        """
        Get the size of the internal cache.

        Returns:
            int: The size of the internal cache.
        """
        return sys.getsizeof(self.internal_memory_map)

    def clean_cache(self):
        """
        Cleans the cache by removing all data except for domain cache.
        """

        self.internal_memory_map = {}



import typing as t
import asyncio
import aiofiles
import datetime
import pathlib

from colorama import Fore
from mailtm.core.methods import ServerAuth, AttachServer
from mailtm.server.cache import InternalCache, CacheType
from mailtm.server.events import (
    BaseEvent,
    NewMessage,
    DomainChange,
    ServerStarted,
    ServerCalledOff,
)
from mailtm.abc.modals import Message, Domain
from mailtm.abc.generic import Token
from mailtm.impls.xclient import AsyncMail

default_banner = pathlib.Path("mailtm/server/assets/banner.txt")
ServerSideEvents = t.TypeVar("ServerSideEvents", bound=BaseEvent)


class MailServerBase:
    """
    Stellar Addition - Custom MailServer.

    This script sets up a [**pooling-based**](https://docs.python.org/3/library/multiprocessing.html) server
    that checks the API every second for new events. When a difference is detected, the corresponding event
    is dispatched, allowing you to respond dynamically to incoming messages.
    In addition to the core SDK functionalities, this package offers an additional layer of scripts designed
    to handle clients in an event-driven manner, reminiscent of frameworks like `discord.py` or `hikari`. With
    this SDK, you gain access to a client that dispatches events seamlessly.

    Parameters
    ----------
    server_auth : ServerAuth
        The server authentication details.
    pooling_rate : Optional[int]
        The pooling rate in seconds. If not provided, the default pooling rate will be used.
    banner : Optional[bool]
        Whether to display a banner upon initialization. Defaults to True.
    banner_path : Optional[Union[Path, str]]
        The path to the banner file. Defaults to the default banner file.
    suppress_errors : Optional[bool]
        Whether to suppress errors. Defaults to False.
    enable_logging : Optional[bool]
        Whether to enable logging. Defaults to False.

    Example
    -------
    ```python
    import mailtm
    server = mailtm.MailServer(
        server_auth=ServerAuth(
            account_id="...",  # Your account ID.
            account_token="...",  # Your account token.
        )
    )
    # Define an event handler for new messages
    # The stand-alone decorators should be used
    # without the brackets since we provide no
    # function to handle, and append with the
    # handler.
    @server.on_new_message
    async def event(event: NewMessage):
        print(event.new_message.text)
    # Start the event loop
    server.run()
    ```
    """

    def __init__(
        self,
        server_auth: ServerAuth,
        pooling_rate: t.Optional[int],
        banner: t.Optional[bool] = True,
        banner_path: t.Optional[t.Union[pathlib.Path, str]] = default_banner,
        suppress_errors: t.Optional[bool] = False,
        enable_logging: t.Optional[bool] = False,
    ) -> None:
        self._banner_enabled = banner
        self._banner_path = banner_path
        self._pooling_rate = pooling_rate
        self._suppress_errors = suppress_errors
        self._logging_enabled = enable_logging
        self.handlers: t.Dict[
            t.Type[BaseEvent], list[t.Callable[[BaseEvent], t.Awaitable[None]]]
        ] = {}
        self._server_auth = server_auth
        self._last_msg: list[Message] = []
        self._last_domain: list[Domain] = []
        self.mail_client = AsyncMail(
            account_token=Token(
                id=self._server_auth.account_id,
                token=self._server_auth.account_token,
            )
        )
        self.collector = InternalCache()

    def _extracted_from_on_new_domain_4(self, arg0: t.Any, func: t.Any):
        if arg0 not in self.handlers:
            self.handlers[arg0] = []
        self.handlers[arg0].append(func)
        return func

    def log(
        self,
        message: str,
        severity: t.Literal["INFO", "WARNING", "ERROR"] = "INFO",
    ) -> None:
        """
        Logs the message to the console with the corresponding severity.

        Parameters
        ----------
        message : str
            The message to log.
        severity : Literal["INFO", "WARNING", "ERROR"]
            The severity of the message. Defaults to INFO.

        Returns
        -------
        None
        """
        if self._logging_enabled is True:
            current_time = datetime.datetime.now().strftime(
                "%a %m/%d/%Y at %I:%M%p"
            )
            if severity == "INFO":
                print(
                    Fore.LIGHTGREEN_EX
                    + f"[+]{Fore.RESET} On {current_time} "
                    + message
                )
            if self._suppress_errors is False:
                if severity == "WARNING":
                    print(
                        Fore.LIGHTYELLOW_EX
                        + f"[!]{Fore.RESET} On {current_time} "
                        + message
                    )
                elif severity == "ERROR":
                    print(
                        Fore.LIGHTRED_EX
                        + f"[-]{Fore.RESET} On {current_time} "
                        + message
                    )
            else:
                print(
                    Fore.LIGHTWHITE_EX
                    + f"[?] Unrecognized severity: {current_time} "
                    + message
                )

    def subscribe(
        self, event_type: t.Type[BaseEvent]
    ) -> t.Callable[
        [t.Callable[[ServerSideEvents], t.Awaitable[None]]],
        t.Callable[[ServerSideEvents], t.Awaitable[None]],
    ]:
        """
        Decorator to subscribe a function to handle server events.

        Parameters
        ----------
        event_type : Type[BaseEvent]
            The type of event to subscribe to.

        Returns
        -------
        Callable
            The decorated function.

        Example
        -------
        ```python
        @server.subscribe(ServerSideEvents.NewMessage)
        async def event(event: ServerSideEvents.NewMessage):
            print(event.new_message.text)
        ```
        """

        def decorator(
            handler_func: t.Callable[[ServerSideEvents], t.Awaitable[None]],
        ) -> t.Callable[[ServerSideEvents], t.Awaitable[None]]:
            if event_type not in self.handlers:
                self.handlers[event_type] = []
            self.handlers[event_type].append(handler_func)  # type: ignore
            return handler_func

        return decorator

    def on_new_message(
        self, func: t.Callable[[NewMessage], t.Awaitable[None]]
    ):
        """
        Registers a callback function to handle new messages.

        Parameters
        ----------
        func : Callable[[NewMessage], Awaitable[None]]
            The callback function to handle new messages.

        Returns
        -------
        The result of the extracted function call.
        """

        return self._extracted_from_on_new_domain_4(NewMessage, func)

    def on_new_domain(
        self, func: t.Callable[[DomainChange], t.Awaitable[None]]
    ):
        """
        Registers a callback function to handle new domains.

        Parameters
        ----------
        func : Callable[[DomainChange], Awaitable[None]]
            The callback function to handle new domains.

        Returns
        -------
        The result of the extracted function call.
        """
        return self._extracted_from_on_new_domain_4(DomainChange, func)

    async def dispatch(self, event: BaseEvent) -> None:
        """
        Asynchronously dispatches the given event to the appropriate handlers.

        Args:
            self: The MailServerBase instance.
            event (BaseEvent): The event to dispatch.

        Returns:
            None
        """
        for handler in self.handlers.get(type(event), []):
            await handler(event)

    async def _check_for_new_messages(self) -> None:
        """
        Checks for new messages by retrieving message information from the mail client.
        If new messages are detected, triggers a NewMessage event with the new message details.
        """
        msg_view = await self.mail_client.get_messages()
        if (
            msg_view
            and msg_view.messages
            and (
                not self._last_msg
                or self._last_msg[0].id != msg_view.messages[0].id
            )
        ):
            if self._last_msg:
                self._last_msg[0] = msg_view.messages[0]
            else:
                self._last_msg.append(msg_view.messages[0])
            new_message_event = NewMessage(
                "NewMessage",
                client=self.mail_client,
                _server=AttachServer(self),
                new_message=msg_view.messages[0],
            )
            await self.dispatch(new_message_event)
            self.collector.add_item_to_cache(
                CacheType.NEW_MESSAGE, msg_view.messages[0]
            )
            self.log(
                message=f"RECEIVED new message from: {msg_view.messages[0].message_from.address}"
            )  # type: ignore
        return None

    async def _check_for_new_domain(self) -> None:
        """
        Asynchronously checks for a new domain by retrieving domain information from the mail client.
        If a new domain is detected, triggers a DomainChange event with the new domain details.
        """
        domain_view = await self.mail_client.get_domains()
        if (
            domain_view
            and domain_view.domains
            and (
                not self._last_domain
                or self._last_domain[0].id != domain_view.domains[0].id
            )
        ):
            if self._last_domain:
                self._last_domain[0] = domain_view.domains[0]
            else:
                self._last_domain.append(domain_view.domains[0])
            new_domain_event = DomainChange(
                event="DomainChange",
                client=self.mail_client,
                _server=AttachServer(self),
                new_domain=domain_view.domains[0],
            )
            await self.dispatch(new_domain_event)
            self.collector.add_item_to_cache(
                CacheType.NEW_MESSAGE, domain_view.domains[0]
            )
            self.log(
                message=f"Domain Changed: {domain_view.domains[0].domain_name}",
                severity="WARNING",
            )

    async def shutdown(self) -> None:
        """
        Calls the shutdown method on the MailClient instance and dispatches the ServerCalledOff event.

        Returns:
            None
        """
        await self.mail_client.close()
        await self.dispatch(
            ServerCalledOff(
                "Server has been called off",
                self.mail_client,
                AttachServer(server=self),
            )
        )
        self.log(
            message="The mail client session has been called off.",
            severity="WARNING",
        )
        loop = asyncio.get_event_loop()
        loop.stop()
        self.log(
            message="Server shuts down on user's request. Goodbye!",
            severity="WARNING",
        )

    async def _banner(self) -> None:
        """
        Asynchronously prints a banner from the specified file path if it exists.
        The banner is formatted with the current time and date, and includes color codes for different parts of the banner.
        If an exception occurs while reading or printing the banner, it is logged.

        :return: None
        """
        if self._banner_path is not None:
            try:
                async with aiofiles.open(
                    self._banner_path, "r", encoding="utf-8"
                ) as f:
                    text = await f.read()
                    details = text.format(
                        time=datetime.datetime.now(),
                        date=f"{datetime.date.today()}",
                        mail=Fore.CYAN,
                        reset=Fore.RESET,
                        sdk=Fore.MAGENTA,
                        ssb=Fore.GREEN,
                        version=Fore.LIGHTBLUE_EX,
                        info=Fore.LIGHTMAGENTA_EX,
                        issues=Fore.RED,
                        warning=Fore.LIGHTYELLOW_EX,
                        dateandtime=Fore.GREEN,
                    )
                    print(details)
            except Exception as e:
                self.log(
                    f"Exception while printing banner:\n{Fore.LIGHTWHITE_EX+str(e)+Fore.RESET}"
                )

    async def runner(self) -> None:
        """
        This function is responsible for executing the main server logic. It starts a session and continuously polls the API for new events. When a difference is detected, the corresponding event is dispatched. The function runs in an infinite loop until it is interrupted by a keyboard interrupt.

        Raises
        ------
            Exception
                If no events have been subscribed to.
        """
        if self._banner_enabled is True:
            await self._banner()
        await self.dispatch(
            ServerStarted(
                "ServerStarted", self.mail_client, AttachServer(self)
            )
        )
        try:
            self.log(
                message="Server-> Session Started: "
                + datetime.datetime.now().strftime("%H:%M:%S")
            )
            self.log(
                message="Server-> Pooling rate: " + str(self._pooling_rate)
            )
            self.log(
                message="Server-> Logging enabled: "
                + str(self._logging_enabled)
            )
            self.log(
                message="Server-> Banner enabled: " + str(self._banner_enabled)
            )
            self.log(
                message="Server-> Subscribed events: "
                + str(len(self.handlers.keys()))
            )
            self.log(
                message="Cache initialized. Created 4 maps.",
                severity="WARNING",
            )
            while True:
                await asyncio.sleep(self._pooling_rate or 1)
                if NewMessage in self.handlers:
                    await self._check_for_new_messages()
                if DomainChange in self.handlers:
                    await self._check_for_new_domain()
                if not self.handlers:
                    await self.mail_client.close()
                    raise RuntimeError(
                        "It seems like you have not subscribed to any events."
                    )
        except Exception as e:
            await self.mail_client.close()
            self.log(
                message=f"Exception while running server:\n{Fore.LIGHTWHITE_EX+str(e)+Fore.RESET}"
            )

    def run(self) -> None:
        """
        Executes the main server logic by running the event runner within asyncio event loop.
        """
        try:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(self.runner())
            loop.close()
        except KeyboardInterrupt:
            asyncio.run(self.shutdown())



"""
🚀 ABCs for implementation the object oriented SDK.

Includes:

- GenericType
    For the generic type used in the SDK.
- ModalType
    For the modals that are used in interaction with API.
"""

__all__ = ["GenericType", "ModalType"]

from . import generic as GenericType
from . import modals as ModalType



"""
- This module provides an asyncio-based client for the Mail.TM API.
- It allows to create, delete, and manage accounts, messages, and sources. As well as retrieve information about domains.
- The API is based on the `aiohttp` library and uses the `msgspec` library to decode the JSON responses from the API into Python objects.
- The client is easy to use and is designed to be used in an asyncio-based application.
- The API is documented at [Mail.tm Documentation](https://docs.mail.tm)

You don't have to worry about:
------------------------------
- The client automatically handles errors and exceptions, so you don't have to.
- The client also handles rate limits, so you don't have to.

You do have to worry about:
---------------------------

- The client does not handle connection limits, so you might have to.
- The client does not handle SSL verification, so you might have to.
"""

from __future__ import annotations

__all__ = ["AsyncMail"]

import aiohttp
import msgspec
import typing as t
import urllib.parse

from ..abc.modals import (
    Account,
    DomainPageView,
    MessagePageView,
    Domain,
    Message,
    Source,
)
from ..abc.generic import Token
from ..core.methods import AccountMethods, DomainMethods, MessageMethods
from ..core.errors import (
    AccountTokenInvalid,
    MissingArgument,
    EntityNotFound,
    MethodNotAllowed,
    RefusedToProcess,
    EntityNotProcessable,
    RatelimitError,
)


class AsyncMail:
    """
    Asynchronous based client handler for the SDK/library.

    Parameters
    ----------
    account_token: Optional[Token]
        The account token to use for authentication. If not provided, the client will not be authenticated.

    Examples
    --------
    ```python
    import asyncio
    from mailtm import AsyncMail

    async def main():
        async with AsyncMail() as client:
            account = await client.get_me()
            print(account)

    asyncio.run(main())
    ```
    """

    def __init__(self, account_token: t.Optional[Token] = None) -> None:
        self._account_token = account_token
        self._base_url = "https://api.mail.tm"
        self._client = aiohttp.ClientSession()
        if self._account_token is not None:
            self._client.headers.update(
                {"Authorization": f"Bearer {self._account_token}"}
            )

    async def _interact(
        self,
        method: t.Literal["GET", "POST", "DELETE", "PATCH"],
        url: str,
        body: t.Optional[t.Any] = None,
        params: t.Optional[t.Dict[str, str]] = None,
    ) -> t.Optional[bytes]:
        """
        Internal method defined to interact with the API using methods, and API slug.

        Parameters
        ----------
        method: Literal["GET", "POST", "DELETE", "PATCH"]
            The method to use to interact with the API.
        url: str
            The API slug to interact with.
        body: Optional[Any]
            The body of the request.
        params: Optional[Dict[str, str]]
            The query parameters of the request.

        Returns
        -------
        Optional[bytes]
            The response from the API.
        """
        if method == "GET":
            result = await self._client.get(url=url, params=params)
        elif method == "POST":
            result = await self._client.post(url=url, json=body)

        elif method == "DELETE":
            result = await self._client.delete(url=url, params=params)

        elif method == "PATCH":
            result = await self._client.patch(url=url, json=body)
        else:
            raise MethodNotAllowed("Report this as a bug on GitHub")

        if str(result.status).startswith("20"):
            return await result.read()
        elif result.status == 400:
            raise MissingArgument(
                "Something in your payload is missing! Or, the payload isn't there at all."
            )
        elif result.status == 401:
            raise AccountTokenInvalid(
                "Your token isn't correct (Or the headers hasn't a token at all!). Remember, every request (Except POST /accounts and POST /token) should be authenticated with a Bearer token!"
            )
        elif result.status == 404:
            raise EntityNotFound(
                "You're trying to access an account that doesn't exist? Or maybe reading a non-existing message? Go check that!"
            )
        elif result.status == 405:
            raise MethodNotAllowed(
                "Maybe you're trying to GET a /token or POST a /messages. Check the path you're trying to make a request to and check if the method is the correct one."
            )
        elif result.status == 418:
            raise RefusedToProcess(
                "Server is a teapot. And refused to process your request at the moment. Kindly contact the developers for further details."
            )
        elif result.status == 422:
            raise EntityNotProcessable(
                "Some went wrong on your payload. Like, the username of the address while creating the account isn't long enough, or, the account's domain isn't correct. Things like that."
            )
        elif result.status == 429:
            raise RatelimitError(
                "You exceeded the limit of 8 requests per second! Try delaying the request by one second!"
            )
        else:
            raise ValueError("Unknown Error")

    async def _create_url(self, other_literal: str) -> str:
        """
        Internal method for creating a URL that joins base with the method slug.

        Parameters
        ----------
        other_literal: str
            The method slug to join with the base URL.

        Returns
        -------
        str
            The joined URL.
        """
        return urllib.parse.urljoin(self._base_url, other_literal)

    async def get_me(self) -> t.Optional[Account]:
        """
        Get the user associated with the account token provided to create a session.

        Returns
        -------
        Optional[Account]
            The user associated with the account token provided to create a session. If not authenticated, returns None.
        """
        resp = await self._interact(
            method="GET", url=await self._create_url(AccountMethods.GET_ME)
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    async def get_domains(self) -> t.Optional[DomainPageView]:
        """
        Get a page view of domains available under the account token provided to create a session.

        Returns
        -------
        Optional[DomainPageView]
            A page view of domains available under the account token provided to create a session. If not authenticated, returns None.
        """
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url, DomainMethods.GET_ALL_DOMAINS
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=DomainPageView, strict=False)
        else:
            return None

    async def get_domain(self, domain_id: str) -> t.Optional[Domain]:
        """
        Get a specific domain with ID.

        Parameters
        ----------
        domain_id: str
            The ID of the domain to get.

        Returns
        -------
        Optional[Domain]
            The domain with the ID provided. If not found, returns None.
        """
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url,
                DomainMethods.GET_DOMAIN_BY_ID.format(id=domain_id),
            ),
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Domain, strict=False)
        else:
            return None

    async def get_account(self, account_id: str) -> t.Optional[Account]:
        """
        Get an accnount by it's ID.

        Parameters
        ----------
        account_id: str
            The ID of the account to get.

        Returns
        -------
        Optional[Account]
            The account with the ID provided. If not found, returns None.
        """
        resp = await self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url,
                AccountMethods.GET_ACCOUNT_BY_ID.format(id=account_id),
            ),
            params={"id": f"{account_id}"},
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    async def create_account(
        self, address: str, password: str
    ) -> t.Optional[Account]:
        """
        Create an account.

        Parameters
        ----------
        address: str
            The email address of the new account.
        password: str
            The password for the new account.

        Returns
        -------
        Optional[Account]
            The newly created account object if successful, None otherwise.
        """
        body = {"address": f"{address}", "password": f"{password}"}
        resp = await self._interact(
            method="POST",
            url=urllib.parse.urljoin(
                self._base_url, AccountMethods.CREATE_ACCOUNT
            ),
            body=body,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Account, strict=False)
        else:
            return None

    async def delete_account(self, account_id: t.Optional[str] = None) -> None:
        """
        Delete an account by it's ID.

        Parameters
        ----------
        account_id: Optional[str]
            The ID of the account to delete. If not provided, the account token will be used.

        Returns
        -------
        None
        """
        if self._account_token is not None and account_id is None:
            await self._interact(
                method="DELETE",
                url=urllib.parse.urljoin(
                    self._base_url,
                    AccountMethods.DELETE_ACCOUNT_BY_ID.format(
                        id=self._account_token.id
                    ),
                ),
            )
        elif account_id is not None:
            await self._interact(
                method="DELETE",
                url=urllib.parse.urljoin(
                    self._base_url,
                    AccountMethods.DELETE_ACCOUNT_BY_ID.format(id=account_id),
                ),
            )
        else:
            raise AccountTokenInvalid(
                "You need an account token to delete an account!"
            )

    async def get_messages(self, page: int = 1) -> t.Optional[MessagePageView]:
        """
        Get a page view of messages available under the account token provided to create a session.

        Parameters
        ----------
        page: int
            The page number to get. Defaults to 1.

        Returns
        -------
        Optional[MessagePageView]
            A page view of messages available under the account token provided to create a session. If not authenticated, returns None.
        """
        params = {"page": f"{page}"}
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url, MessageMethods.GET_ALL_MESSAGES
            ),
            params=params,
        )
        if resp is not None:
            return msgspec.json.decode(
                resp, type=MessagePageView, strict=False
            )
        else:
            return None

    async def get_message(self, message_id: str) -> t.Optional[Message]:
        """
        Get a specific message with ID.

        Parameters
        ----------
        message_id: str
            The ID of the message to get.

        Returns
        -------
        Optional[Message]
            The message with the ID provided. If not found, returns None.
        """
        params = {"id": f"{message_id}"}
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url,
                MessageMethods.GET_MESSAGE_BY_ID.format(id=message_id),
            ),
            params=params,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Message, strict=False)
        else:
            return None

    async def delete_message(self, message_id: str) -> None:
        """
        Delete a specific message with ID.

        Parameters
        ----------
        message_id: str
            The ID of the message to delete.

        Returns
        -------
        None
        """
        params = {"id": f"{message_id}"}
        await self._interact(
            method="DELETE",
            url=urllib.parse.urljoin(
                self._base_url,
                MessageMethods.DELETE_MESSAGE_BY_ID.format(id=message_id),
            ),
            params=params,
        )

    async def mark_as_seen(self, message_id: str) -> None:
        """
        Flag a message as seen.

        Parameters
        ----------
        message_id: str
            The ID of the message to mark as seen.

        Returns
        -------
        None
        """
        params = {"id": f"{message_id}"}
        await self._interact(
            method="PATCH",
            url=urllib.parse.urljoin(
                self._base_url,
                MessageMethods.PATCH_MESSAGE_BY_ID.format(id=message_id),
            ),
            params=params,
        )

    async def get_source(self, source_id: str) -> t.Optional[Source]:
        """
        Get source by the source ID.

        Parameters
        ----------
        source_id: str
            The ID of the source to get.

        Returns
        -------
        Optional[Source]
            The source with the ID provided. If not found, returns None.
        """
        params = {"id": f"{source_id}"}
        resp = await self._interact(
            method="GET",
            url=urllib.parse.urljoin(
                self._base_url,
                MessageMethods.GET_SOURCES_BY_ID.format(id=source_id),
            ),
            params=params,
        )
        if resp is not None:
            return msgspec.json.decode(resp, type=Source, strict=False)
        else:
            return None

    async def close(self):
        """
        Close the client.
        """
        await self._client.close()



"""
🚀 Methods that are used in the API, to make calls to the API.

Includes:

- errors: Exceptions that are used in the API.
- methods: Methods that are used in the API.

"""

__all__ = ["errors", "methods"]

from . import errors
from . import methods



"""
🚀 This package not only encompasses fundamental SDK functionalities but also integrates an additional
layer of scripts tailored specifically for managing clients in an event-driven fashion, reminiscent of
popular frameworks such as discord.py or hikari. By utilizing this SDK, you gain access to a client that
seamlessly dispatches events, enhancing the overall functionality and versatility of your application.

Includes:

- ⚡ Mail Server
- ⚡ Server Events
"""

__all__ = ["MailServer", "ServerEvents"]

from .impl import MailServer
from . import events as ServerEvents



import msgspec
import datetime
import typing as t


from .generic import (
    MessageFrom,
    MessageTo,
    MessageAttachment,
    ViewDetails,
    ViewSearch,
)


class Domain(msgspec.Struct):
    """
    The domain of the email account.
    """

    _id: t.Optional[str] = msgspec.field(name="@id", default=None)
    """
    `Secrative`: ID of the interaction.
    
    Returns
    -------
    id : str
        ID of the interaction.
    """

    _type: t.Optional[str] = msgspec.field(name="@type", default=None)
    """
    `Secrative`: Type of the interaction.
    
    Returns
    -------
    type : str
        Type of the interaction.
    """

    _context: t.Optional[str] = msgspec.field(name="@context", default=None)
    """
    `Secrative`: Context of the interaction.
    
    Returns
    -------
    context : str
        Context of the interaction.
    """

    id: t.Optional[str] = msgspec.field(name="id", default=None)
    """
    `Not documented`: ID of the interaction.
    
    Returns
    -------
    id : str
        ID of the interaction.
    """
    domain_name: t.Optional[str] = msgspec.field(name="domain", default=None)
    """
    Name of the domain provided by mail.tm

    Returns
    -------
    domain : str
        Name of the domain provided by mail.tm

    Examples
    --------
    @gmail.com, @goster.com
    """
    is_active: t.Optional[bool] = msgspec.field(name="isActive", default=None)
    """
    If the domain is still active.
    
    Returns
    -------
    is_active : bool
        If the domain is still active.
    """

    is_private: t.Optional[bool] = msgspec.field(
        name="isPrivate", default=None
    )
    """
    If the domain is private. Private domains are not visible to the public.
    
    Returns
    -------
    is_private : bool
        If the domain is private. Private domains are not visible to the public.
    """
    created_at: t.Optional[datetime.datetime] = msgspec.field(
        name="createdAt", default=None
    )
    """
    The datetime object of creation date of the domain.
    
    Returns
    -------
    created_at : datetime
        The datetime object of creation date of the domain.
    """

    updated_at: t.Optional[datetime.datetime] = msgspec.field(
        name="updatedAt", default=None
    )
    """
    The datetime object of update date of the domain from the latest point of reference.

    Returns
    -------
    updated_at : datetime
        The datetime object of update date of the domain from the latest point of reference.
    """


class Account(msgspec.Struct):
    _id: t.Optional[str] = msgspec.field(name="@id", default=None)
    """
    `Secrative`: ID of the account.
    
    Returns
    -------
    id : str
        ID of the account.
    """

    _type: t.Optional[str] = msgspec.field(name="@type", default=None)
    """
    `Secrative`: Type of the account.
    
    Returns
    -------
    type : str
        Type of the account.
    """

    _context: t.Optional[str] = msgspec.field(name="@context", default=None)
    """
    `Secrative`: Context of the account.
    
    Returns
    -------
    context : str
        Context of the account.
    """

    id: t.Optional[str] = msgspec.field(name="id", default=None)
    """
    `Not documented`: ID of the account.
    
    Returns
    -------
    id : str
        ID of the account.
    """

    address: t.Optional[str] = msgspec.field(name="address", default=None)
    """
    Email address of the account.
    Returns
    -------
    address : str
        Email address of the account.
    """

    quota: t.Optional[int] = msgspec.field(name="quota", default=None)
    """
    The quota of the account.
    
    Returns
    -------
    quota : int
        The quota of the account.
    """

    used: t.Optional[int] = msgspec.field(name="used", default=None)
    """
    The amount of quota used by the account.
    
    Returns
    -------
    used : int
        The amount of quota used by the account.
    """

    is_disabled: t.Optional[bool] = msgspec.field(
        name="isDisabled", default=None
    )
    """
    If the account is disabled.
    
    Returns
    -------
    is_disabled : bool
        If the account is disabled.
    """

    is_deleted: t.Optional[bool] = msgspec.field(
        name="isDeleted", default=None
    )
    """
    If the account is deleted.
    
    Returns
    -------
    is_deleted : bool
        If the account is deleted.
    """

    created_at: t.Optional[datetime.datetime] = msgspec.field(
        name="createdAt", default=None
    )
    """
    The datetime object of creation date of the account.
    
    Returns
    -------
    created_at : datetime
        The datetime object of creation date of the account.
    """

    updated_at: t.Optional[datetime.datetime] = msgspec.field(
        name="updatedAt", default=None
    )
    """
    The datetime object of update date of the account from the latest point of reference.
    
    Returns
    -------
    updated_at : datetime
        The datetime object of update date of the account from the latest point of reference.
    """


class Message(msgspec.Struct):
    _id: t.Optional[str] = msgspec.field(name="@id", default=None)
    """
    `Secrative`: ID of the message.
    
    Returns
    -------
    id : str
        ID of the message.
    """

    _type: t.Optional[str] = msgspec.field(name="@type", default=None)
    """
    `Secrative`: Type of the message.
    
    Returns
    -------
    type : str
        Type of the message.
    """

    _context: t.Optional[str] = msgspec.field(name="@context", default=None)
    """
    `Secrative`: Context of the message.
    
    Returns
    -------
    context : str
        Context of the message.
    """

    id: t.Optional[str] = msgspec.field(name="id", default=None)
    """
    ID of the message.
    
    Returns
    -------
    id : str
        ID of the message.
    """

    account_id: t.Optional[str] = msgspec.field(name="accountId", default=None)
    """
    ID of the account to which the message belongs.
    
    Returns
    -------
    account_id : str
        ID of the account to which the message belongs.
    """

    message_id: t.Optional[str] = msgspec.field(name="msgid", default=None)
    """
    The ID associated with the message.
    
    Returns
    -------
    message_id : str
        Message ID.
    """

    message_from: t.Optional[MessageFrom] = msgspec.field(
        name="from", default=None
    )
    """
    Details of the sender of the message.
    
    Returns
    -------
    message_from : MessageFrom
        Details of the sender of the message.
    """

    message_to: t.Optional[t.List[MessageTo]] = msgspec.field(
        name="to", default=None
    )
    """
    Details of the recipients of the message.
    
    Returns
    -------
    message_to : MessageTo
        Details of the recipients of the message.
    """

    subject: t.Optional[str] = msgspec.field(name="subject", default=None)
    """
    Subject of the message.
    
    Returns
    -------
    subject : str
        Subject of the message.
    """

    seen: t.Optional[bool] = msgspec.field(name="seen", default=None)
    """
    If the message has been seen by the recipient.
    
    Returns
    -------
    seen : bool
        If the message has been seen by the recipient.
    """

    is_deleted: t.Optional[bool] = msgspec.field(
        name="isDeleted", default=None
    )
    """
    If the message is deleted.
    
    Returns
    -------
    is_deleted : bool
        If the message is deleted.
    """

    html: t.Optional[t.List[str]] = msgspec.field(name="html", default=None)
    """
    HTML content of the message.
    
    Returns
    -------
    html : str
        HTML content of the message.
    """

    has_attachments: t.Optional[bool] = msgspec.field(
        name="hasAttachments", default=None
    )
    """
    If the message has attachments.
    
    Returns
    -------
    has_attachments : bool
        If the message has attachments.
    """

    attachments: t.Optional[t.List[MessageAttachment]] = msgspec.field(
        name="attachments", default=None
    )
    """
    Attachments associated with the message.
    
    Returns
    -------
    attachments : MessageAttachment
        Attachments associated with the message.
    """

    size: t.Optional[int] = msgspec.field(name="size", default=None)
    """
    Size of the message in bytes.
    
    Returns
    -------
    size : int
        Size of the message in bytes.
    """

    downloadUrl: t.Optional[str] = msgspec.field(
        name="downloadUrl", default=None
    )
    """
    URL to download the message.
    
    Returns
    -------
    downloadUrl : str
        URL to download the message.
    """

    created_at: t.Optional[datetime.datetime] = msgspec.field(
        name="createdAt", default=None
    )
    """
    Date and time of creation of the message.
    
    Returns
    -------
    created_at : datetime
        Date and time of creation of the message.
    """

    updated_at: t.Optional[datetime.datetime] = msgspec.field(
        name="updatedAt", default=None
    )
    """
    Date and time of last update of the message.
    
    Returns
    -------
    updated_at : datetime
        Date and time of last update of the message.
    """

    cc: t.Optional[t.List[str]] = msgspec.field(name="cc", default=None)
    """
    Carbon Copy (CC) recipients of the message.
    
    Returns
    -------
    cc : str
        Carbon Copy (CC) recipients of the message.
    """

    bcc: t.Optional[t.List[str]] = msgspec.field(name="bcc", default=None)
    """
    Blind Carbon Copy (BCC) recipients of the message.
    
    Returns
    -------
    bcc : str
        Blind Carbon Copy (BCC) recipients of the message.
    """

    flagged: t.Optional[bool] = msgspec.field(name="flagged", default=None)
    """
    If the message is flagged by the recipient.
    
    Returns
    -------
    flagged : bool
        If the message is flagged by the recipient.
    """

    verifications: t.Optional[t.List[str]] = msgspec.field(
        name="verifications", default=None
    )
    """
    Verifications associated with the message.
    
    Returns
    -------
    verifications : str
        Verifications associated with the message.
    """

    retention_date: t.Optional[datetime.datetime] = msgspec.field(
        name="retentionDate", default=None
    )
    """
    Date of retention for the message.
    
    Returns
    -------
    retention_date : datetime
        Date of retention for the message.
    """

    retention: t.Optional[bool] = msgspec.field(name="retention", default=None)
    """
    If the message is subject to retention.
    
    Returns
    -------
    retention : bool
        If the message is subject to retention.
    """

    text: t.Optional[str] = msgspec.field(name="text", default=None)
    """
    Plain text content of the message.
    
    Returns
    -------
    text : str
        Plain text content of the message.
    """


class Source:
    _id: t.Optional[str] = msgspec.field(name="@id", default=None)
    """
    `Secrative`: ID of the source.
    
    Returns
    -------
    id : str
        ID of the source.
    """

    _type: t.Optional[str] = msgspec.field(name="@type", default=None)
    """
    `Secrative`: Type of the source.
    
    Returns
    -------
    type : str
        Type of the source.
    """

    _context: t.Optional[str] = msgspec.field(name="@context", default=None)
    """
    `Secrative`: Context of the source.
    
    Returns
    -------
    context : str
        Context of the source.
    """

    id: t.Optional[str] = msgspec.field(name="id", default=None)
    """
    The id attribute of the Source.
    
    Returns
    -------
    id : str
        The id attribute of the Source.
    """

    download_url: t.Optional[str] = msgspec.field(
        name="downloadUrl", default=None
    )
    """
    The download URL attribute of the Source.
    
    Returns
    -------
    download_url : str
        The download URL attribute of the Source.
    """

    data: t.Optional[str] = msgspec.field(name="data", default=None)
    """
    The data attribute of the Source.
    
    Returns
    -------
    data : str
        The data attribute of the Source.
    """


class MessagePageView(msgspec.Struct):
    """
    Page view for messages under a page.
    """

    messages: t.Optional[t.List[Message]] = msgspec.field(
        name="hydra:member", default=None
    )
    """
    List of messages in the view.
    
    Returns
    -------
    messages : Message
        List of messages in the view.
    """

    total_items: t.Optional[int] = msgspec.field(
        name="hydra:totalItems", default=None
    )
    """
    Total number of items in the view.
    
    Returns
    -------
    total_items : int
        Total number of items in the view.
    """

    view_search: t.Optional[ViewSearch] = msgspec.field(
        name="hydra:search", default=None
    )
    """
    Search parameters of the view.
    
    Returns
    -------
    view_search : ViewSearch
        Search parameters of the view.
    """

    view_details: t.Optional[ViewDetails] = msgspec.field(
        name="hydra:view", default=None
    )
    """
    Details of the view.
    
    Returns
    -------
    view_details : ViewDetails
        Details of the view.
    """


class DomainPageView(msgspec.Struct):
    """
    Page view for domains under a page.
    """

    domains: t.Optional[t.List[Domain]] = msgspec.field(
        name="hydra:member", default=None
    )
    """
    List of domains in the view.
    
    Returns
    -------
    domains : Domain
        List of domains in the view.
    """

    total_items: t.Optional[int] = msgspec.field(
        name="hydra:totalItems", default=None
    )
    """
    Total number of domains in the view.
    
    Returns
    -------
    total_items : int
        Total number of domains in the view.
    """

    view_details: t.Optional[ViewDetails] = msgspec.field(
        name="hydra:view", default=None
    )
    """
    Details of the domain view.
    
    Returns
    -------
    view_details : ViewDetails
        Details of the domain view.
    """

    view_search: t.Optional[ViewSearch] = msgspec.field(
        name="hydra:search", default=None
    )
    """
    Search parameters of the domain view.
    
    Returns
    -------
    view_search : ViewSearch
        Search parameters of the domain view.
    """



import pathlib
import typing as t

from .events import (
    NewAccountCreated,
    MessageDelete,
    AccountSwitched,
    AccountDeleted,
)
from .srv import MailServerBase, default_banner
from .cache import CacheType

from ..core.methods import AttachServer, ServerAuth
from ..abc.modals import Account
from ..abc.generic import Token
from ..impls.pullers import xget


class MailServer(MailServerBase):
    def __init__(
        self,
        server_auth: ServerAuth,
        pooling_rate: int | None,
        banner: bool | None = True,
        banner_path: pathlib.Path | str | None = default_banner,
        suppress_errors: bool | None = False,
        enable_logging: bool | None = False,
    ) -> None:
        self.pull = xget()
        super().__init__(
            server_auth,
            pooling_rate,
            banner,
            banner_path,
            suppress_errors,
            enable_logging,
        )

    async def create_account(
        self, account_address: str, account_password: str
    ) -> t.Optional[Account]:
        """
        Creates a new account with the given email address and password.

        Parameters
        ----------
        account_address : str
            The email address of the new account.
        account_password: str
            The password for the new account.

        Returns
        -------
        Optional[Account]
            The newly created account object if successful, None otherwise.
        """
        try:
            new_account = await self.mail_client.create_account(
                account_address, account_password
            )
            token = await self.pull.get_account_token(
                account_address, account_password
            )
            if token is not None and new_account is not None:
                account_auth = ServerAuth(
                    account_token=token.token,
                    account_id=token.id,
                    account_address=account_address,
                    account_password=account_password,
                )
                await self.dispatch(
                    NewAccountCreated(
                        new_account=new_account,
                        new_account_auth=account_auth,
                        event="A new account got created.",
                        client=self.mail_client,
                        _server=AttachServer(self),
                    )
                )
                self.collector.add_item_to_cache(
                    cache_type=CacheType.NEW_ACCOUNTS, item=new_account
                )
            return new_account
        except Exception as e:
            self.log(
                message="Could not create account: " + str(e), severity="ERROR"
            )
            return None

    async def delete_message(self, message_id: str) -> None:
        """
        Deletes a message from the Mail Box.

        Parameters
        ----------
        message_id : str
            The ID of the message to be deleted.

        Returns
        -------
            None
        """
        message = await self.mail_client.get_message(message_id=message_id)
        if message:
            try:
                await self.mail_client.delete_message(message_id)
                await self.dispatch(
                    MessageDelete(
                        "Deleted a message.",
                        deleted_message=message,
                        client=self.mail_client,
                        _server=AttachServer(self),
                    )
                )
                self.collector.add_item_to_cache(
                    cache_type=CacheType.OLD_MESSAGE, item=message
                )
            except Exception as e:
                self.log(
                    message="Could not delete message: " + str(e),
                    severity="ERROR",
                )
                return None

    async def switch_account(
        self, new_account_token: t.Union[Token, str]
    ) -> None:
        """
        Switches to a new account.

        Parameters
        ----------
        new_account_token : t.Union[Token, str]
            The new account token to switch to.

        Returns
        -------
            None
        """
        try:
            if isinstance(new_account_token, Token):
                self.mail_client._client.headers.update(
                    {"Authorization": f"Bearer {new_account_token.token}"}
                )  # type: ignore
                self.log(
                    message=f"Switched to new account with ID: {new_account_token.id}",
                    severity="WARNING",
                )
            if isinstance(new_account_token, str):
                self.mail_client._client.headers.update(
                    {"Authorization": f"Bearer {new_account_token}"}
                )  # type: ignore
                self.log(
                    message=f"Switched to new account with Token: {new_account_token}",
                    severity="WARNING",
                )

            await self.dispatch(
                AccountSwitched(
                    event="Account switched",
                    client=self.mail_client,
                    _server=AttachServer(self),
                    last_account_auth=self._server_auth,
                )
            )

        except Exception as e:
            self.log(
                message="Could not switch account: " + str(e), severity="ERROR"
            )

    async def delete_account(self, account_id: str) -> None:
        """
        Deletes an account.

        Parameters
        ----------
        account_id : str
            The ID of the account to be deleted.

        Returns
        -------
            None
        """
        try:
            await self.mail_client.delete_account(account_id)
            self.log(message="Deleted account with ID: " + account_id)
            await self.dispatch(
                AccountDeleted(
                    event=f"Account has been deleted with the ID: {account_id}",
                    client=self.mail_client,
                    _server=AttachServer(self),
                )
            )
        except Exception as e:
            self.log(
                message="Could not delete account: " + str(e), severity="ERROR"
            )

    async def shutdown(self) -> None:
        self.collector.clean_cache()
        return await super().shutdown()

    async def runner(self) -> None:
        self.collector.build_cache()
        return await super().runner()



class MissingArgument(Exception):
    """
    Bad request 400

    Something in your payload is missing! Or, the payload isn't there at all.
    """

    ...


class AccountTokenInvalid(Exception):
    """
    Unauthorized 401

    Your token isn't correct (Or the headers hasn't a token at all!). Remember, every request (Except POST /accounts and POST /token) should be authenticated with a Bearer token!
    """

    ...


class EntityNotFound(Exception):
    """
    Not found 404

    You're trying to access an account that doesn't exist? Or maybe reading a non-existing message? Go check that!
    """

    ...


class MethodNotAllowed(Exception):
    """
    Method not allowed 405

    Maybe you're trying to GET a /token or POST a /messages. Check the path you're trying to make a request to and check if the method is the correct one.
    """

    ...


class RefusedToProcess(Exception):
    """
    I'm a teapot 418

    Who knows? Maybe the server becomes a teapot!
    """

    ...


class EntityNotProcessable(Exception):
    """
    Unprocessable entity 422

    Something went wrong on your payload. Like, the username of the address while creating the account isn't long enough, or, the account's domain isn't correct. Things like that.
    """

    ...


class RatelimitError(Exception):
    """
    Too many requests 429

    You exceeded the limit of 8 requests per second! Try delaying the request by one second!
    """

    ...



"""
🚀 Implementations for clients and pullers.

These includes:
- `clients` - `AsyncMail` for asychronous operations and `SyncMail` for synchronous operations.
- `pullers` - Pullers are components within a system that retrieve data or messages from a source in a controlled and efficient manner.
"""

__all__ = ["AsyncMail", "SyncMail", "xget", "get"]

from .client import SyncMail
from .xclient import AsyncMail
from .pullers import get, xget



from __future__ import annotations
from setuptools import setup, find_namespace_packages

with open("README.md", "r", encoding="utf-8") as f:
    description = f.read()


setup(
    name="mailtm-sdk",
    version="0.1.3",
    description="Mail.tm Stack Development Kit, designed to enhance your experience with the renowned temporary email service, mail.tm",
    long_description=description,
    long_description_content_type="text/markdown",
    author="Parth Mishra",
    author_email="halfstackpgr@gmail.com",
    maintainer="Parth Mishra",
    maintainer_email="halfstackpgr@gmail.com",
    url="https://halfstackpgr.github.io/Mail.tm/",
    packages=find_namespace_packages(include=["mailtm*"]),
    entry_points={"console_scripts": ["mailtm = mailtm.cli:version"]},
    install_requires=["aiohttp", "requests", "msgspec", "colorama"],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
)



"""
This module contains classes representing different methods available in the Mail.tm webservice.
Each class represents a set of related methods.
The classes are used to construct method calls to the API of Mail.tm and make requests to their server.
"""

from __future__ import annotations

__all__ = [
    "DomainMethods",
    "AccountMethods",
    "MessageMethods",
    "ServerAuth",
    "AttachServer",
]

from dataclasses import dataclass
import typing as t

if t.TYPE_CHECKING:
    from mailtm.server.srv import MailServerBase


@dataclass
class DomainMethods:
    """
    Represents methods related to domains in Mail.tm webservice.
    """

    GET_ALL_DOMAINS = "domains"
    """
    Retrieves all domains.
    """
    GET_DOMAIN_BY_ID = "domains/{id}"
    """
    Retrieves a domain by ID.
    """


@dataclass
class AccountMethods:
    """
    Represents methods related to accounts in Mail.tm webservice.
    """

    CREATE_ACCOUNT = "accounts"
    """
    Creates a new account.
    """
    GET_ACCOUNT_BY_ID = "accounts/{id}"
    """
    Retrieves an account by ID.
    """
    DELETE_ACCOUNT_BY_ID = "accounts/{id}"
    """
    Deletes an account by ID.
    """
    GET_ME = "me"
    """
    Retrieves the current logged in account.
    """

    GET_ACCOUNT_TOKEN = "token"
    """
    Retrieves the token of the account.
    """


@dataclass
class MessageMethods:
    """
    Represents methods related to messages in Mail.tm webservice.
    """

    GET_ALL_MESSAGES = "messages"
    """
    Retrieves all messages.
    """
    GET_MESSAGE_BY_ID = "messages/{id}"
    """
    Retrieves a message by ID.
    """

    DELETE_MESSAGE_BY_ID = "messages/{id}"
    """
    Deletes a message by ID.
    """

    PATCH_MESSAGE_BY_ID = "messages/{id}"
    """
    Updates a message by ID.
    """

    GET_SOURCES_BY_ID = "sources/{id}"

    """
    Retrieves sources for a message by ID.
    """


@dataclass
class ServerAuth:
    """
    A data class to authenticate with the server-client for Server Implementation.
    """

    account_token: str
    """
    The token of the account.
    """
    account_id: str
    """
    The ID of the account.
    """
    account_address: t.Optional[str] = None
    """
    The address of the account.
    """
    account_password: t.Optional[str] = None
    """
    The password of the account."""


@dataclass
class AttachServer:
    """
    This contains an internal server instance to perform methods inherited from the server.
    """

    server: MailServerBase



"""
## 📨 Mail.tm Stack Development Kit

Welcome to the Mail.tm Stack Development Kit, designed to enhance your experience with
the renowned temporary email service, [Mail.tm](http://mail.tm). Here's why you'll find
it indispensable:

1. #### Enhanced Integration:
This kit streamlines interactions with Mail.tm's API, empowering you to seamlessly
integrate its features into your applications. Whether you're automating email
workflows or incorporating temporary email capabilities, this SDK has you covered.

2. #### Comprehensive Documentation: Methods are constructed using the documentation,
available [here](http://docs.mail.tm).

3. #### Adherence to Terms:
Rest assured, this SDK complies fully with Mail.tm's terms of usage, ensuring a seamless
and secure experience for both developers and users. By respecting Mail.tm's conditions,
we prioritize the integrity and reliability of the service by this kit.

4. #### Assurance of safety:
This repository prioritizes your safety. With frequent updates to
dependencies, we ensure that no vulnerable dependencies compromise your security. Count on us
for a secure and reliable experience.

Experience the power and convenience of Mail.tm with our Stack Development Kit - your gateway to efficient and reliable temporary email solutions.

---
### Credits:
Created by GitHub: [@halfstackpgr](https://github.com/halfstackpgr)
For any query or bug: [Raise an Issues](https://github.com/halfstackpgr/mailtm/issues)
For further updates visit: [GitHub](https://github.com/halfstackpgr/mailtm)
"""

__version__ = "0.1.3"

__all__ = [
    "MailServer",
    "ServerEvents",
    "GenericTypes",
    "ModalTypes",
    "AsyncMail",
    "SyncMail",
    "xget",
    "get",
    "errors",
    "methods",
]

from .server import events as ServerEvents
from .server.impl import MailServer
from .abc import generic as GenericTypes
from .abc import modals as ModalTypes
from .impls.xclient import AsyncMail
from .impls.client import SyncMail
from .impls.pullers import get, xget
from .core import errors, methods



"""
Module for events that get triggered in the Mail.tm server.

This module contains a set of classes representing different events that get triggered
in the `Mail.tm` server. These events can be used to create callbacks for when certain
actions happen in the server.

- `NewMessage`: Triggered when a new message is received.
- `MessageDelete`: Triggered when a message is deleted using the server instance.
- `DomainChange`: Triggered when the email domain is changed.
- `AccountSwitched`: Triggered when the account is switched to a different account.
- `NewAccountCreated`: Triggered when a new account is created.
- `AccountDeleted`: Triggered when an account is deleted.
- `ServerStarted`: Triggered when the server is started.
- `ServerCalledOff`: Triggered when the server is ended.
"""

__all__ = [
    "NewMessage",
    "MessageDelete",
    "DomainChange",
    "AccountSwitched",
    "NewAccountCreated",
    "AccountDeleted",
    "ServerStarted",
    "ServerCalledOff",
]


from ..impls.xclient import AsyncMail
from ..core.methods import AttachServer
from ..abc.modals import Message, Domain, Account
from ..core.methods import ServerAuth


class BaseEvent:
    """
    Represents a set of Discord UI components attached to a message.

    Parameters
    ----------
    event : str
        A string attached to the BaseEvent that represents the event.
    client : AsyncMail
        An instance AsyncMail that represents the client. You can use this to operate the interactions with the AsyncClient

    _server: AttachServer
        This is not supposed to be used by the user. This attaches a server instance to dispatch events within from events.
    """

    def __init__(
        self, event: str, client: AsyncMail, _server: AttachServer
    ) -> None:
        self._server = _server
        self._event = event
        self._client = client

    @property
    def client(self) -> AsyncMail:
        """
        Client attached to the server context.

        Returns
        -------
        AsyncMail
            an instance of AsyncMail attached to the server in the context of the interaction.
        """
        return self._client

    @property
    def server(self) -> AttachServer:
        """
        An instance of AttachServer, which basically has the MailServer attached to it.

        Returns
        -------
        AttachServer
            an instance of AttachServer, which can be used like `AttachServer.server`. This contains `MailServer`.
        """
        return self._server

    @property
    def event(self) -> str:
        """
        The event string associated with the dispatched class.

        Returns
        -------
        str
            The event string associated with the dispatched class
        """
        return self._event


class NewMessage(BaseEvent):
    """
    Event triggered when a message is received.

    Parameters
    ----------
    new_message : Message
        The message that was received.
    """

    def __init__(
        self,
        event: str,
        client: AsyncMail,
        _server: AttachServer,
        new_message: Message,
    ) -> None:
        self._new_message = new_message
        super().__init__(event, client, _server)

    @property
    def new_message(self) -> Message:
        """
        The new message that was received from the server.

        Returns
        -------
        Message
            An instance of Message which includes the details about message received from the server.
        """
        return self.new_message

    async def delete_message(self) -> None:
        """
        Delete the message from the Mail Box.
        """
        if self.new_message.message_id is not None:
            await self._server.server.dispatch(
                MessageDelete(
                    "MessageDelete",
                    self.client,
                    self.new_message,
                    self._server,
                )
            )
            await self.client.delete_message(self.new_message.message_id)

    async def mark_as_seen(self) -> None:
        """
        Flag the message as seen.
        """
        if self.new_message.message_id is not None:
            await self.client.mark_as_seen(self.new_message.message_id)


class MessageDelete(BaseEvent):
    """
    Event triggered when a message  gets deleted using the server instance.
    Note: This won't get triggered if the client is used to delete any message

    Parameters
    ----------
    deleted_message : Message
        The message that was deleted.
    """

    def __init__(
        self,
        event: str,
        client: AsyncMail,
        deleted_message: Message,
        _server: AttachServer,
    ) -> None:
        self._deleted_message = deleted_message
        super().__init__(event, client, _server)

    @property
    def deleted_message(self) -> Message:
        """
        The message that was deleted.

        Returns
        -------
        Message
            The message that was deleted.
        """
        return self._deleted_message


class DomainChange(BaseEvent):
    """
    Event triggered when the email domain is changed.

    Parameter
    ---------
    new_domain : Domain
        The new domain that was set.
    """

    def __init__(
        self,
        new_domain: Domain,
        event: str,
        client: AsyncMail,
        _server: AttachServer,
    ) -> None:
        self._new_domain = new_domain
        super().__init__(event, client, _server)

    @property
    def new_domain(self) -> Domain:
        """
        The domain that got changed.

        Returns
        -------
        Domain
            A domain that was changed
        """
        return self._new_domain


class AccountSwitched(BaseEvent):
    """
    Event triggered when the account is switched to a different account.

    Parameter
    ---------
    last_account_auth: ServerAuth
        An instance of ServerAuth that represents the last account.

    """

    def __init__(
        self,
        event: str,
        client: AsyncMail,
        _server: AttachServer,
        last_account_auth: ServerAuth,
    ) -> None:
        self._last_account_auth = last_account_auth
        super().__init__(event, client, _server)

    @property
    def last_account_auth(self) -> ServerAuth:
        """
        Get the ServerAuth of the last account.

        Returns:
        --------
        ServerAuth
            An instance of ServerAuth that represents the last account.
        """
        return self.last_account_auth


class NewAccountCreated(BaseEvent):
    """
    Event triggered when a new account is created.

    Parameters
    ----------
    new_account_auth: ServerAuth
        An instance of ServerAuth that represents the new account.

    new_account: Account
        An instance of Account that represents the new account.
    """

    def __init__(
        self,
        new_account_auth: ServerAuth,
        new_account: Account,
        event: str,
        client: AsyncMail,
        _server: AttachServer,
    ) -> None:
        self._new_account_aut = new_account_auth
        self._new_account = new_account
        super().__init__(event, client, _server)

    @property
    def new_account_auth(self) -> ServerAuth:
        """
        The server authentication related to the new account.

        Returns
        -------
        ServerAuth:
            An instance representing the details used in server authentication.
        """
        return self._new_account_aut

    @property
    def new_account(self) -> Account:
        """
        The new account that was created.

        Returns
        -------
        Account
            An instance of Account that represents the new account.
        """
        return self._new_account


class AccountDeleted(BaseEvent):
    """
    Event triggered when an account is deleted.
    """

    def __init__(
        self, event: str, client: AsyncMail, _server: AttachServer
    ) -> None:
        super().__init__(event, client, _server)


class ServerStarted(BaseEvent):
    """
    Event triggered when the server is started.
    """

    def __init__(
        self, event: str, client: AsyncMail, _server: AttachServer
    ) -> None:
        super().__init__(event, client, _server)


class ServerCalledOff(BaseEvent):
    """
    Event triggered when the server is ended.
    """

    def __init__(
        self, event: str, client: AsyncMail, _server: AttachServer
    ) -> None:
        super().__init__(event, client, _server)



from colorama import Fore
import datetime

text = r"""
{mail} __  __       _ _  _{reset}                           {ssb}Server-Side Build{reset}
{mail}|  \/  |     (_) || |{reset}
{mail}| \  / | __ _ _| || |_ _ __ ___{reset}    {info}Developed: halfstackpgr     Website: https://mail.tm{reset}
{mail}| |\/| |/ _` | | || __| '_ ` _ \{reset}                        
{mail}| |  | | (_| | | || |_| | | | | |{reset}  {info}Documentation: https://github.com/halfstackpgr/Mail.tm{reset}
{mail}|_|  |_|\__,_|_|_(_)__|_| |_| |_|{reset}                       
        {sdk} ____  ____  _  __{reset}         {issues}Issues: https://github.com/halfstackpgr/Mail.tm/Issues{reset}
        {sdk}/ ___||  _ \| |/ /{reset}        
        {sdk}\___ \| | | | ' /{reset}          {version}Version: 0.1.3                            Coverage: 98%{reset}  
        {sdk} ___) | |_| | . \{reset}       
        {sdk}|____/|____/|_|\_\{reset}         {dateandtime}Time: {time}                             Date:{date}{reset}
"""


def version() -> None:
    """
    Version information and general banner for Mail.TM.
    """
    details = text.format(
        time=datetime.datetime.now().time().strftime("%H:%M:%S"),
        date=f"{datetime.date.today()}",
        mail=Fore.CYAN,
        reset=Fore.RESET,
        sdk=Fore.MAGENTA,
        ssb=Fore.GREEN,
        version=Fore.LIGHTBLUE_EX,
        info=Fore.LIGHTMAGENTA_EX,
        issues=Fore.RED,
        warning=Fore.LIGHTYELLOW_EX,
        dateandtime=Fore.GREEN,
    )

    print(details)



"""
This module provides the following classes for type-safe and msgspec-based handling of the Mail.tm API data types.

- `MessageFrom`: Represents a data class containing details of messages.
- `MessageTo`: Represents a data class containing details of recipients.
- `MessageAttachment`: Represents a data class containing details of attachments.
- `ViewDetails`: Represents a data class containing details of search results.
- `ViewSearch`: Represents a data class containing details of search queries.

These classes are designed to be used with the 'msgspec' library to facilitate quick and efficient data serialization and deserialization.
"""

from __future__ import annotations

__all__ = [
    "MessageFrom",
    "MessageTo",
    "MessageAttachment",
    "Token",
    "ViewDetails",
    "ViewMapping",
    "ViewSearch",
]

import msgspec
import typing as t


class MessageFrom(msgspec.Struct):
    """
    Represents a data class containing details of messages.
    """

    name: str = msgspec.field(name="name")
    """
    Name of the `Account` by which the `Message` was sent.
    
    Returns
    -------
    name : str
        Name of the account message is sent from.
    """
    address: str = msgspec.field(name="address")
    """
    Email address of the `Account` by which the `Message` was sent.
    
    Returns
    -------
    address : str
        Email address of the account message is sent from.
    """


class MessageTo(msgspec.Struct):
    name: str = msgspec.field(name="name")
    """
    Name of the `Account` to which the `Message` was sent.
    
    Returns
    -------
    name : str
        Name of the account message is sent to.
    """
    address: str = msgspec.field(name="address")
    """
    Email address of the `Account` to which the `Message` was sent.
    
    Returns
    -------
    address : str
        Email address of the account message is sent to.
    """


class MessageAttachment(msgspec.Struct):
    id: str = msgspec.field(name="id")
    """
    ID of the message attachment.
    
    Returns
    -------
    id : str
        ID of the message attachment.
    """
    filename: str = msgspec.field(name="filename")
    """
    The name of the attachment file.
    
    Returns
    -------
    filename : str
        The name of the attachment file.
    """
    content_type: str = msgspec.field(name="contentType")
    """
    The MIME type of the attachment.
    
    Returns
    -------
    content_type : str
        The MIME type of the attachment.
    """
    disposition: str = msgspec.field(name="disposition")
    """
    The Content-Disposition header of the attachment.
    
    Returns
    -------
    disposition : str
        The Content-Disposition header of the attachment.
    """
    transfer_encoding: str = msgspec.field(name="transferEncoding")
    """
    The Transfer-Encoding header of the attachment.
    
    Returns
    -------
    transfer_encoding : str
        The Transfer-Encoding header of the attachment.
    """
    related: bool = msgspec.field(name="related")
    """
    Whether the attachment is related to the main body of the message.
    
    Returns
    -------
    related : bool
        Whether the attachment is related to the main body of the message.
    """
    size: int = msgspec.field(name="size")
    """
    The size of the attachment in bytes.
    
    Returns
    -------
    size : int
        The size of the attachment in bytes.
    """
    download_url: str = msgspec.field(name="downloadUrl")
    """
    The URL where the attachment can be downloaded from.
    
    Returns
    -------
    download_url : str
        The URL where the attachment can be downloaded from.
    """


class Token(msgspec.Struct):
    id: str = msgspec.field(name="id")
    """
    ID of the account.
    
    Returns
    -------
    id : str
        ID of the account.
    """
    token: str = msgspec.field(name="token")
    """
    Token of the account.
    
    Returns
    -------
    token : str
        Token of the account.
    """

    def __str__(self) -> str:
        return self.token

    def __repr__(self) -> str:
        return self.__str__()


class ViewDetails(msgspec.Struct):
    """
    Struct representing the details of a view.
    """

    _id: str = msgspec.field(name="@id")
    """
    The unique identifier of the view.
    
    Returns
    -------
    id : str
        The unique identifier of the view.
    """
    _type: str = msgspec.field(name="@type")
    """
    The type of the view.
    
    Returns
    -------
    type : str
        The type of the view.
    """
    first: str = msgspec.field(name="hydra:first")
    """
    The URL of the first page in the view.
    
    Returns
    -------
    first : str
        The URL of the first page in the view.
    """
    last: str = msgspec.field(name="hydra:last")
    """
    The URL of the last page in the view.
    
    Returns
    -------
    last : str
        The URL of the last page in the view.
    """
    previous: str = msgspec.field(name="hydra:previous")
    """
    The URL of the previous page in the view.
    
    Returns
    -------
    previous : str
        The URL of the previous page in the view.
    """
    next: str = msgspec.field(name="hydra:next")
    """
    The URL of the next page in the view.
    
    Returns
    -------
    next : str
        The URL of the next page in the view.
    """


class ViewMapping(msgspec.Struct):
    """
    Struct representing a mapping between a variable and a property.
    """

    _type: str = msgspec.field(name="@type")
    """
    The type of the mapping.
    
    Returns
    -------
    type : str
        The type of the mapping.
    """
    variable: str = msgspec.field(name="variable")
    """
    The variable of the mapping.
    
    Returns
    -------
    variable : str
        The variable of the mapping.
    """
    property: str = msgspec.field(name="property")
    """
    The property of the mapping.
    
    Returns
    -------
    property : str
        The property of the mapping.
    """
    required: bool = msgspec.field(name="required")
    """
    Whether the mapping is required.
    
    Returns
    -------
    required : bool
        Whether the mapping is required.
    """


class ViewSearch(msgspec.Struct):
    """
    A view search system.
    """

    _type: str = msgspec.field(name="@type")
    """
    The type of the view.
    
    Returns
    -------
    type : str
        The type of the view.
    """

    template: str = msgspec.field(name="hydra:template")
    """
    The URL template of the view.
    
    Returns
    -------
    template : str
        The URL template of the view.
    """

    variable_representation: str = msgspec.field(
        name="hydra:variableRepresentation"
    )
    """
    The representation of variables in the view.
    
    Returns
    -------
    variable_representation : str
        The representation of variables in the view.
    """

    mappings: t.List[ViewMapping] = msgspec.field(name="hydra:mapping")
    """
    A list of mappings for the view.
    
    Returns
    -------
    mappings : List[ViewMapping]
        A list of mappings for the view.
    """

    def __str__(self) -> str:
        return self.template

    def __repr__(self) -> str:
        return self.__str__()
